!--------1---------2---------3---------4---------5---------6---------7---------8
! 
!  Module   hardwareinfo_mod
!> @brief   utilities for get hardware information
!! @authors Chigusa Kobayashi (CK)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
! 
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module hardwareinfo_mod

  use fileio_str_mod
  use fileio_mod
  use string_mod
  use constants_mod
  use messages_mod


  implicit none
  private

  character(13), parameter  :: cpuinfo = "/proc/cpuinfo"

  public  :: hw_information
  private :: runtime_information
  private :: build_information
  private :: genesis_information
  private :: get_cpu_information
  private :: get_env_information

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    hw_information
  !> @brief        write architecture and compile information
  !! @param[in]    get_arch_infomation 
  !! @authors      CK
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine hw_information

    call genesis_information

    call build_information

    call runtime_information

    return

  end subroutine hw_information
  
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    runtime_information
  !> @brief        write runtime information
  !! @authors      CK
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine runtime_information
    character(MaxLine)       :: ldlibrary,  rt_user, rt_host
    character(80)            :: cpuname
    integer                  :: tval(8)
    integer(8)               :: year, month, day, hour, minute, sec

    write(MsgOut,'(A)') 'Runtime_Information> Machine and Library Information'

    call date_and_time(values=tval)
    year   = int(tval(1),kind=8)
    month  = int(tval(2),kind=8)
    day    = int(tval(3),kind=8)
    hour   = int(tval(5),kind=8)
    minute = int(tval(6),kind=8)
    sec    = int(tval(7),kind=8)
    write(MsgOut,'(A,i4.4,a,i2.2,a,i2.2,a,i2.2,a,i2.2,a,i2.2)') &
    '  date       = ',&
      year,"/",month,"/",day," ",hour,":",minute,":", sec

    call get_cpu_information(cpuname)
    write(MsgOut,*) ' cpu model    = ',trim(cpuname)
    call get_env_information(ldlibrary,rt_host,rt_user)
    write(MsgOut,*) ' exec. host   = ',trim(rt_user),'@',trim(rt_host)
    write(MsgOut,*) ' LD library   = ',trim(ldlibrary)
    write(MsgOut,'(A)') 

    return

  end subroutine runtime_information

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    genesis_information
  !> @brief        write GENESIS information
  !! @authors      CK
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine genesis_information

    write(MsgOut,'(A)') 'GENESIS_Information> GENESIS Information'
    write(MsgOut,*) ' version      = ',VERSION
    write(MsgOut,*) ' commit ID    = ',COMPILE_GENESIS_VERSION
    write(MsgOut,*) ' precision    = ',precision_char
#ifdef USE_GPU
    write(MsgOut,*) ' nonbonding   = GPU'
#else
    write(MsgOut,*) ' nonbonding   = CPU'
#endif
    write(MsgOut,'(A)') 

    return

  end subroutine genesis_information


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    build_information
  !> @brief        write build information
  !! @authors      CK
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine build_information

    write(MsgOut,'(A)') 'Build_Information> Compiler Information'
    write(MsgOut,*) ' build host   = ',COMPILE_USER,'@',COMPILE_HOST
    write(MsgOut,*) ' fortran      = ',COMPILE_FC_VER
    write(MsgOut,*) ' option       = ',COMPILE_FCFLAGS
    write(MsgOut,*) ' C            = ',COMPILE_CC_VER
    write(MsgOut,*) ' option       = ',COMPILE_CFLAGS
    write(MsgOut,*) ' defined var. = ',COMPILE_DEFINED_VARIABLES
    write(MsgOut,*) ' link option  = ',COMPILE_LDFLAGS
#ifdef CUDAGPU
    write(MsgOut,*) ' CUDA         = ',COMPILE_NVCC_VER
#endif
    write(MsgOut,'(A)') 

    return

  end subroutine build_information


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    get_env_information
  !> @brief        write Enviroment information
  !! @authors      CK
  !! @param[out]   ldlibrary    : LD_LIBRARY_PATH
  !! @param[out]   rt_host      : host
  !! @param[out]   rt_user      : user
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine get_env_information(ldlibrary, rt_host, rt_user)

#if defined(INTEL)
    use ifport, only:getenv

#elif defined(KCOMP) || defined(RICC)
    use service_routines, only:getenv

#endif

    character(*),            intent(out)    :: ldlibrary
    character(*),            intent(out)    :: rt_host
    character(*),            intent(out)    :: rt_user

    call getenv("LD_LIBRARY_PATH",ldlibrary)
    call getenv("USER",rt_user)
    call getenv("HOSTNAME",rt_host)

    return

  end subroutine get_env_information

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    get_cpu_information
  !> @brief        write CPU information
  !! @param[out]   cpuname    :  cpu model name
  !! @authors      CK
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine get_cpu_information(cpuname)

    character(*),            intent(out)    :: cpuname

    integer       :: i, inchar, file
    character(80) :: line
    logical       :: exists

    exists=.false.
    
    inquire(file=cpuinfo, EXIST=exists)
    if (.not. exists) then
      cpuname="N/A"
      return
    endif

    call open_file(file, cpuinfo, IOFileInput)

    do while(.true.)
      read(file, '(A80)',end=100,err=100) line
      inchar = index(line,":")
      if (inchar > 0) then
        if (line(1:10) == 'model name') then
           read(line(inchar+2:80),'(A)') cpuname
           exit
        else if (inchar == 6 .and. line(1:3) == 'cpu') then
          read(line(inchar+2:80),'(A)') cpuname
          exit
        endif
      endif
    end do
100 continue

    call close_file(file)

    return

  end subroutine get_cpu_information

end module hardwareinfo_mod
